# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------
# pylint: disable=line-too-long

from knack.arguments import CLIArgumentType

from azure.cli.core.commands.parameters import (resource_group_name_type, get_location_type,
                                                get_resource_name_completion_list,
                                                get_three_state_flag, get_enum_type, tags_type)

def load_arguments(self, _):

    name_type = CLIArgumentType(options_list=['--name', '-n'])

    with self.argument_context('containerapp') as c:
        # Base arguments
        c.argument('name', name_type, metavar='NAME', id_part='name')
        c.argument('resource_group_name', arg_type=resource_group_name_type)
        c.argument('location', arg_type=get_location_type(self.cli_ctx))

    with self.argument_context('containerapp create') as c:
        c.argument('tags', tags_type)
        c.argument('kube_env', options_list=['--environment', '-e'], help="Name of the containerapp's environment.")
        c.argument('yaml', help='Path to a .yaml file with the configuration of a containerapp. All other parameters will be ignored')

    # Container
    with self.argument_context('containerapp create', arg_group='Container') as c:
        c.argument('image_name', options_list=['--image', '-i'], help="Container image, e.g. publisher/image-name:tag. If there are multiple containers, please use --yaml instead.")
        c.argument('cpu', options_list=['--cpu'], help="Required CPU in cores, e.g. 0.5")
        c.argument('memory', options_list=['--memory'], help="Required memory, e.g. 250Mi")
        c.argument('env_vars', options_list=['--environment-variables', '-v'], help="A list of environment variable(s) for the containerapp. Comma-separated values in 'key=value' format. If there are multiple containers, please use --yaml instead.")
        c.argument('startup_command', options_list=['--command'], help="A list of supported commands on the container app that will executed during container startup. Comma-separated values e.g. '/bin/queue'. If there are multiple containers, please use --yaml instead.")
        c.argument('args', options_list=['--args'], help="A list of container startup command argument(s). Comma-separated values e.g. '-c, mycommand'. If there are multiple containers, please use --yaml instead.")

    # Scale
    with self.argument_context('containerapp create', arg_group='Scale') as c:
        c.argument('min_replicas', options_list=['--min-replicas'], help="The minimum number of containerapp replicas.")
        c.argument('max_replicas', options_list=['--max-replicas'], help="The maximum number of containerapp replicas.")
        c.argument('scale_rules', options_list=['--scale-rules'], help="The name of a yaml file containing a list of scale rules.")

    # Configuration
    with self.argument_context('containerapp create', arg_group='Configuration') as c:
        c.argument('revisions_mode', arg_type=get_enum_type(['single', 'multiple']), options_list=['--revisions-mode'], help="The active revisions mode for the containerapp.")
        c.argument('registry_server', options_list=['--registry-login-server'], help="The url of the registry, e.g. myregistry.azurecr.io")
        c.argument('registry_pass', options_list=['--registry-password'], help="The password to log in container image registry server")
        c.argument('registry_user', options_list=['--registry-username'], help="The username to log in container image registry server")
        c.argument('secrets', options_list=['--secrets', '-s'], help="A list of secret(s) for the containerapp. Comma-separated values in 'key=value' format.")

    # Ingress
    with self.argument_context('containerapp create', arg_group='Ingress') as c:
        c.argument('ingress', options_list=['--ingress'], default=None, arg_type=get_enum_type(['internal', 'external']), help="Ingress type that allows either internal or external+internal ingress traffic to the Containerapp.")
        c.argument('transport', arg_type=get_enum_type(['auto', 'http', 'http2']), help="The transport protocol used for ingress traffic.")

    # Dapr
    with self.argument_context('containerapp create', arg_group='Dapr') as c:
        c.argument('dapr_enabled', options_list=['--enable-dapr'], default=False, arg_type=get_three_state_flag())
        c.argument('dapr_app_port', options_list=['--dapr-app-port'], help="The port of your app.")
        c.argument('dapr_app_id', options_list=['--dapr-app-id'], help="The dapr app id.")
        c.argument('dapr_components', options_list=['--dapr-components'], help="The name of a yaml file containing a list of dapr components.")
    
    with self.argument_context('containerapp update') as c:
        c.argument('tags', tags_type)
        c.argument('yaml', help='Path to a .yaml file with the configuration of a containerapp. All other parameters will be ignored')

    # Container
    with self.argument_context('containerapp update', arg_group='Container') as c:
        c.argument('image_name', options_list=['--image', '-i'], help="Container image name, e.g. publisher/image-name:tag. If there are multiple containers, please use --yaml instead.")
        c.argument('cpu', options_list=['--cpu'], help="Required CPU in cores, e.g. 0.5")
        c.argument('memory', options_list=['--memory'], help="Required memory, e.g. 250Mi")
        c.argument('env_vars', options_list=['--environment-variables', '-v'], help="A list of environment variable(s) for the containerapp. Comma-separated values in 'key=value' format. If there are multiple containers, please use --yaml instead.")
        c.argument('startup_command', options_list=['--command'], help="A list of supported commands on the container app that will executed during container startup. Comma-separated values e.g. '/bin/queue'. If there are multiple containers, please use --yaml instead.")
        c.argument('args', options_list=['--args'], help="A list of container startup command argument(s). Comma-separated values e.g. '-c, mycommand'. If there are multiple containers, please use --yaml instead.")

    # Scale
    with self.argument_context('containerapp update', arg_group='Scale') as c:
        c.argument('min_replicas', options_list=['--min-replicas'], help="The minimum number of containerapp replicas.")
        c.argument('max_replicas', options_list=['--max-replicas'], help="The maximum number of containerapp replicas.")
        c.argument('scale_rules', options_list=['--scale-rules'], help="The name of a yaml file containing a list of scale rules.")
        
    # Configuration
    with self.argument_context('containerapp update', arg_group='Configuration') as c:
        c.argument('revisions_mode', arg_type=get_enum_type(['single', 'multiple']), options_list=['--revisions-mode'], help="The active revisions mode for the containerapp.")
        c.argument('registry_server', options_list=['--registry-login-server'], help="The url of the registry, e.g. myregistry.azurecr.io")
        c.argument('registry_pass', options_list=['--registry-password'], help="The password to log in container image registry server")
        c.argument('registry_user', options_list=['--registry-username'], help="The username to log in container image registry server")
        c.argument('secrets', options_list=['--secrets', '-s'], help="A list of secret(s) for the containerapp. Comma-separated values in 'key=value' format.")

    # Ingress
    with self.argument_context('containerapp update', arg_group='Ingress') as c:
        c.argument('ingress', options_list=['--ingress'], default=None, arg_type=get_enum_type(['internal', 'external', 'none']), help="Ingress type that allows either internal or external+internal ingress traffic to the Containerapp, or \'none\' to remove.")
        c.argument('target_port', options_list=['--target-port'], help="The application port used for ingress traffic.")
        c.argument('transport', arg_type=get_enum_type(['auto', 'http', 'http2']), help="The transport protocol used for ingress traffic.")
        c.argument('traffic_weights', options_list=['--traffic-weight'], help="A list of revision weight(s) for the containerapp. Comma-separated values in 'revision_name=weight' format.")

    # Dapr
    with self.argument_context('containerapp update', arg_group='Dapr') as c:
        c.argument('dapr_enabled', options_list=['--enable-dapr'], default=False, arg_type=get_three_state_flag())
        c.argument('dapr_app_port', options_list=['--dapr-app-port'], help="The port of your app.")
        c.argument('dapr_app_id', options_list=['--dapr-app-id'], help="The dapr app id.")
        c.argument('dapr_components', options_list=['--dapr-components'], help="The name of a yaml file containing a list of dapr components.")

    with self.argument_context('containerapp revision deactivate') as c:
        c.argument('name', arg_type=name_type)
        c.argument('app', options_list=['--app'], help="Name of the containerapp")
    
    with self.argument_context('containerapp revision activate') as c:
        c.argument('name', arg_type=name_type)
        c.argument('app', options_list=['--app'], help="Name of the containerapp")

    with self.argument_context('containerapp revision restart') as c:
        c.argument('name', arg_type=name_type)
        c.argument('app', options_list=['--app'], help="Name of the containerapp")

    with self.argument_context('containerapp revision show') as c:
        c.argument('name', arg_type=name_type)
        c.argument('app', options_list=['--app'], help="Name of the containerapp")

    with self.argument_context('containerapp show') as c:
        c.argument('name', arg_type=name_type)

    with self.argument_context('containerapp secret') as c:
        c.argument('secret_name', options_list=['--secret-name'])

    with self.argument_context('containerapp secret set') as c:
        c.argument('value', options_list=['--value'])

    with self.argument_context('containerapp env set') as c:
        c.argument('value', options_list=['--value'])
        c.argument('secret_ref', options_list=['--secret-ref'])

    with self.argument_context('containerapp port') as c:
        c.argument('port_name', options_list=['--port-name'])

    with self.argument_context('containerapp port add') as c:
        c.argument('port', options_list=['--port'])
        c.argument('protocol', options_list=['--protocol'])

    with self.argument_context('containerapp port update') as c:
        c.argument('port', options_list=['--port'])
        c.argument('protocol', options_list=['--protocol'])

    with self.argument_context('containerapp scale') as c:
        c.argument('min_replicas', options_list=['--min-replicas'], help='the minimum number of replicas for the container app')
        c.argument('max_replicas', options_list=['--max-replicas'], help='the maximum number of replicas for the container app')

    with self.argument_context('containerapp scale-rule') as c:
        c.argument('rule_name', options_list=['--rule-name'])

    with self.argument_context('containerapp scale-rule create') as c:
        c.argument('rule_type', options_list=['--rule-type'], help='type of scale rule. One of azure-queue, custom or http')
        c.argument('queue_name', options_list=['--queue-name'], help='for type: azure-queue. Name of the Azure Queue')
        c.argument('queue_length', options_list=['--queue-length'], help='for type: azure-queue. Length of the Azure Queue')
        c.argument('metadata', options_list=['--metadata'], help='for type: http or custom. properties for the http scale rule')
        c.argument('custom_type', options_list=['--custom-type'], help='for type: custom. type of custom type')

    with self.argument_context('containerapp scale-rule update') as c:
        c.argument('queue_name', options_list=['--queue-name'], help='name of the Azure Queue')
        c.argument('queue_length', options_list=['--queue-length'], help='length of the Azure Queue')

    with self.argument_context('containerapp env') as c:
        c.argument('name', name_type)
        c.argument('resource_group_name', arg_type=resource_group_name_type)
        c.argument('location', arg_type=get_location_type(self.cli_ctx), help='Location of resource. Examples: Canada Central, North Europe')
        c.argument('logs_destination', options_list=['--logs-dest'])
        c.argument('logs_customer_id', options_list=['--logs-workspace-id'])
        c.argument('logs_key', options_list=['--logs-workspace-key'])
        c.argument('instrumentation_key', options_list=['--instrumentation-key'])
        c.argument('controlplane_subnet_resource_id', options_list=['--controlplane-subnet-resource-id'], help='Resource ID of a subnet for control plane infrastructure components. This subnet must be in the same VNET as the subnet defined in appSubnetResourceId.')
        c.argument('app_subnet_resource_id', options_list=['--app-subnet-resource-id'], help='Resource ID of a subnet that Container App containers are injected into. This subnet must be in the same VNET as the subnet defined in controlPlaneSubnetResourceId.')
        c.argument('tags', arg_type=tags_type)

    with self.argument_context('containerapp env update') as c:
        c.argument('name', name_type, help='Name of the kubernetes environment.')
        c.argument('tags', arg_type=tags_type)
    
    with self.argument_context('containerapp env delete') as c:
        c.argument('name', name_type, help='Name of the Kubernetes Environment.')

    with self.argument_context('containerapp env show') as c:
        c.argument('name', name_type, help='Name of the Kubernetes Environment.')

    with self.argument_context('containerapp env wait') as c:
        c.argument('name', name_type, help='Name of the Kubernetes Environment.')

    with self.argument_context('containerapp traffic') as c:
        c.argument('name', name_type)
        c.argument('weights', options_list=['--revision-weight'], help="A list of revision weight(s) for the containerapp. Comma-separated values in 'key=value' format.")

    with self.argument_context('containerapp github-action add') as c:
        c.argument('repo_url', help='The GitHub repository to which the workflow file will be added. In the format: https://github.com/<owner>/<repository-name>')
        c.argument('token', help='A Personal Access Token with write access to the specified repository. For more information: https://help.github.com/en/github/authenticating-to-github/creating-a-personal-access-token-for-the-command-line')
        c.argument('branch', options_list=['--branch', '-b'], help='The branch of the GitHub repo. Defaults to "master" if not specified.')
        c.argument('login_with_github', help='Interactively log in with Github to retrieve the Personal Access Token')
        c.argument('registry_url', help='The url of the registry, e.g. myregistry.azurecr.io')
        c.argument('registry_username', help='The username of the registry. If using Azure Container Registry, we will try to infer the credentials if not supplied')
        c.argument('registry_password', help='The password of the registry. If using Azure Container Registry, we will try to infer the credentials if not supplied')
        c.argument('docker_file_path', help='The dockerfile location, e.g. ./Dockerfile')
        c.argument('service_principal_client_id', help='The service principal client ID. ')
        c.argument('service_principal_client_secret', help='The service principal client secret.')
        c.argument('service_principal_tenant_id', help='The service principal tenant ID.')

    with self.argument_context('containerapp github-action delete') as c:
        c.argument('token', help='A Personal Access Token with write access to the specified repository. For more information: https://help.github.com/en/github/authenticating-to-github/creating-a-personal-access-token-for-the-command-line')
        c.argument('login_with_github', help='Interactively log in with Github to retrieve the Personal Access Token')
