import yaml
from knack.util import CLIError
from azure.cli.core.azclierror import (ResourceNotFoundError, RequiredArgumentMissingError, MutuallyExclusiveArgumentError, ValidationError)
from ._client_factory import cf_resource_groups

def dapr_components_from_file(cmd, file):
    DaprComponent, DaprMetadata = cmd.get_models('DaprComponent', 'DaprMetadata')
    dapr_components_def = []
    
    with open(file, 'r') as stream:
        data = yaml.safe_load(stream)

        if not isinstance(data, list):
           raise CLIError("--dapr-components: must be an array of dapr component(s)")

        for dapr in data:
            if dapr.get("type") is None:
                raise CLIError("--dapr-components: type must be specified for dapr component")
            if dapr.get("version") is None:
                raise CLIError("--dapr-components: version must be specified for dapr component")
            if dapr.get("name") is None:
                raise CLIError("--dapr-components: name must be specified for dapr component")

            component = DaprComponent(name=dapr.get("name"), version=dapr.get("version"), type=dapr.get("type"))

            if dapr.get("metadata") is not None:
                component_metadata = []

                if not isinstance(data, list):
                    raise CLIError("--dapr-components: metadata be an array of name value pairs")

                for meta in dapr.get("metadata"):
                    if meta.get("name") is None:
                        raise CLIError("--dapr-components: metadata name must be specified")
                    if meta.get("value") is None and meta.get("secretRef") is None:
                        raise CLIError("--dapr-components: value or secretRef must be specified")
                    elif meta.get("value") is not None:
                        component_metadata.append(DaprMetadata(name=meta.get("name"), value=meta.get("value")))
                    elif meta.get("secretRef") is not None:
                        component_metadata.append(DaprMetadata(name=meta.get("name"), secret_ref=meta.get("secretRef")))
                
                component.metadata = component_metadata

            dapr_components_def.append(component)
            
    return dapr_components_def

def scale_rules_from_file(cmd, file):
    CustomScaleRule, ScaleRule, ScaleRuleAuth = cmd.get_models('CustomScaleRule', 'ScaleRule', 'ScaleRuleAuth')
    workerapp_scale_rules_def = []

    with open(file, 'r') as stream:
        data = yaml.safe_load(stream)

        if not isinstance(data, list):
           raise CLIError("--scale-rules: must be an array of scale rule(s)")
        
        for rule in data:
            if rule.get("name") is None:
                raise CLIError("--scale-rules: name field required in rules config")
            if rule.get("type") is None:
                raise CLIError("--scale-rules: type field required in rules config")
            
            if rule.get("type") == "http":
                http_scale_rule = CustomScaleRule(metadata=rule.get("metadata"), type="http")
                workerapp_scale_rules_def.append(ScaleRule(name=rule.get("name"), custom=http_scale_rule))

    return workerapp_scale_rules_def

def contains(list, filter):
    for x in list:
        if filter(x):
            return True
    return False

def parse_list_of_strings(comma_separated_string):
    comma_separated = comma_separated_string.split(',')
    return [s.strip() for s in comma_separated]

def combine_secrets(existing_secrets, new_secrets):
    existing_secrets = existing_secrets or []
    new_secrets = new_secrets or []

    secrets = new_secrets
    secrets_set = {}

    for i in range(len(new_secrets)):
        if new_secrets[i].name in secrets_set:
            raise CLIError("Duplicate secret {secret} found, secret names must be unique.".format(secret =new_secrets[i].name))
        secrets_set[new_secrets[i].name] = True

    for i in range(len(existing_secrets)):
        if existing_secrets[i].name not in secrets_set:
            secrets.append(existing_secrets[i])
    return [secret for secret in secrets if secret.value]

def combine_env_vars(existing_env_vars, new_env_vars):
    existing_env_vars = existing_env_vars or []
    new_env_vars = new_env_vars or []

    env_vars = existing_env_vars
    dict_existing_env_vars = {}

    for i in range(len(existing_env_vars)):
        dict_existing_env_vars[existing_env_vars[i].name] = i

    for i in range(len(new_env_vars)):
        if new_env_vars[i].name in dict_existing_env_vars:
            env_vars[dict_existing_env_vars[new_env_vars[i].name]] = new_env_vars[i]
        else:
            env_vars.append(new_env_vars[i])

    env_vars = [env_var for env_var in env_vars if (env_var.value or env_var.secret_ref)]

    return env_vars

def parse_env_var_flags(cmd, env_string, is_update_containerapp=False):
    # Useful Resources
    EnvironmentVar = cmd.get_models('EnvironmentVar')

    env_pair_strings = env_string.split(',')
    env_pairs = {}

    for pair in env_pair_strings:
        key_val = pair.split('=')
        if len(key_val) is not 2:
            if is_update_containerapp:
                raise CLIError("Environment variables must be in the format \"<key>=<value>,<key>=secretref:<value>,...\". If you are updating a Containerapp, did you pass in the flag \"--environment\"? Updating a containerapp environment is not supported, please re-run the command without this flag.")
            raise CLIError("Environment variables must be in the format \"<key>=<value>,<key>=secretref:<value>,...\".")
        if key_val[0] in env_pairs:
            raise CLIError("Duplicate environment variable {env} found, environment variable names must be unique.".format(env = key_val[0]))
        value = key_val[1].split('secretref:')
        env_pairs[key_val[0]] = value

    env_var_def = []
    for key, value in env_pairs.items():
        if len(value) is 2:
            env_var_def.append(EnvironmentVar(name=key, secret_ref=value[1]))
        else:
            env_var_def.append(EnvironmentVar(name=key, value=value[0]))

    return env_var_def

def parse_secret_flags(cmd, secret_string):
    # Useful Resources
    Secret = cmd.get_models('Secret')

    secret_pair_strings = secret_string.split(',')
    secret_pairs = {}

    for pair in secret_pair_strings:
        key_val = pair.split('=', 1)
        if len(key_val) is not 2:
            raise CLIError("--secrets: must be in format \"<key>=<value>,<key>=<value>,...\"")
        if key_val[0] in secret_pairs:
            raise CLIError("--secrets: duplicate secret {secret} found, secret names must be unique.".format(secret = key_val[0]))
        secret_pairs[key_val[0]] = key_val[1]

    secret_var_def = []
    for key, value in secret_pairs.items():
        secret_var_def.append(Secret(name=key, value=value))

    return secret_var_def

def parse_traffic_weights_flags(cmd, weight_string):
    # Useful Resources
    TrafficWeight = cmd.get_models('TrafficWeight')

    weight_pair_strings = weight_string.split(',')
    weight_pairs = {}

    for pair in weight_pair_strings:
        key_val = pair.split('=', 1)
        if len(key_val) is not 2:
            raise CLIError("--traffic-weight: must be in format \"<revision>=<weight>,<revision>=<weight>,...\"")
        if key_val[0] in weight_pairs:
            raise CLIError("--traffic-weight: duplicate revision {revision} found, revisions must be unique.".format(revision = key_val[0]))
        if is_valid_weight(key_val[1]):
            weight_pairs[key_val[0]] = key_val[1]
        else:
            raise CLIError("--traffic-weight: weight for revision {revision} must be an integer between 0 and 100".format(revision = key_val[0]))
    
    traffic_weight_def = []
    for key, value in weight_pairs.items():
        if key == 'latest':
            traffic_weight_def.append(TrafficWeight(latest_revision=True, weight=value))
        else:
            traffic_weight_def.append(TrafficWeight(revision_name=key, weight=value))

    return traffic_weight_def

def is_number(s):
    try:
        float(s)
        return True
    except ValueError:
        return False

def validate_memory(memory):
    valid = False

    if memory.endswith("Gi"):
        valid = is_number(memory[:-2])

    if not valid:
        raise ValidationError("usage error: --memory must be a number ending with \"Gi\"")

def validate_cpu(cpu):
    if not is_number(cpu):
        raise ValidationError("usage error: --cpu must be a number eg. \"0.5\"")

def is_valid_weight(s):
    try: 
        n = int(s)
        if n >= 0 and n <= 100:
            return True
        return False
    except ValueError:
        return False

def validate_transport_method(method):
    return method == "auto" or method == "http" or method == "http2"

def _get_location_from_resource_group(cli_ctx, resource_group_name):
    client = cf_resource_groups(cli_ctx)
    group = client.get(resource_group_name)
    return group.location

def raise_missing_token_suggestion():
    pat_documentation = "https://help.github.com/en/articles/creating-a-personal-access-token-for-the-command-line"
    raise RequiredArgumentMissingError("GitHub access token is required to authenticate to your repositories. "
                                       "If you need to create a Github Personal Access Token, "
                                       "please run with the '--login-with-github' flag or follow "
                                       "the steps found at the following link:\n{0}".format(pat_documentation))