# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# pylint: disable=line-too-long
from azure.cli.core.commands import CliCommandType
from msrestazure.tools import is_valid_resource_id, parse_resource_id
from ._client_factory import ex_handler_factory, cf_containerapps, cf_containerapps_revisions, cf_kube_environments

def transform_containerapp_output(app):
    props = ['name', 'location', 'resourceGroup', 'provisioningState']
    result = {k: app[k] for k in app if k in props}

    if 'latestRevisionFqdn' in app:
        result['fqdn'] = app['latestRevisionFqdn']

    return result

def transform_containerapp_list_output(apps):
    return [transform_containerapp_output(a) for a in apps]

def transform_revision_output(rev):
    props = ['name', 'replicas', 'active', 'createdTime']
    result = {k: rev[k] for k in rev if k in props}

    if 'latestRevisionFqdn' in rev['template']:
        result['fqdn'] = rev['template']['latestRevisionFqdn']

    return result

def transform_revision_list_output(revs):
    return [transform_revision_output(r) for r in revs]


def transform_containerapp_create_or_update_output(containerapp):
    # Don't display kube environment ID since it's too long for table output
    result = {k: containerapp[k] for k in containerapp if k != 'kubeEnvironmentId'}

    if 'kubeEnvironmentId' in containerapp and is_valid_resource_id(containerapp['kubeEnvironmentId']):
        parsed_kube_env = parse_resource_id(containerapp['kubeEnvironmentId'])
        kube_name = parsed_kube_env['name']
        result['kubeEnvironment'] = kube_name
    return result


def load_command_table(self, _):

    containerapps_sdk = CliCommandType(
        operations_tmpl='azext_workerapp.vendored_sdks.azure_mgmt_web.operations#ContainerAppsOperations.{}',
        client_factory=cf_containerapps
    )

    containerapps_revisions_sdk = CliCommandType(
        operations_tmpl='azext_workerapp.vendored_sdks.azure_mgmt_web.operations#ContainerAppsRevisionsOperations.{}',
        client_factory=cf_containerapps_revisions
    )

    kube_environments_sdk = CliCommandType(
        operations_tmpl='azext_workerapp.vendored_sdks.azure_mgmt_web.operations#KubeEnvironmentsOperations.{}',
        client_factory=cf_kube_environments
    )

    with self.command_group('containerapp', containerapps_sdk, is_preview=True) as g:
        g.custom_command('create', 'create_containerapp', client_factory=cf_containerapps, supports_no_wait=True, exception_handler=ex_handler_factory(), table_transformer=transform_containerapp_create_or_update_output)
        g.custom_command('delete', 'delete_containerapp', client_factory=cf_containerapps, supports_no_wait=True, confirmation=True, exception_handler=ex_handler_factory())
        g.custom_command('list', 'list_containerapp', client_factory=cf_containerapps, table_transformer=transform_containerapp_list_output)
        g.custom_command('scale', 'scale_containerapp', client_factory=cf_containerapps, exception_handler=ex_handler_factory())
        g.custom_command('update', 'update_containerapp', client_factory=cf_containerapps, exception_handler=ex_handler_factory(), table_transformer=transform_containerapp_create_or_update_output)
        g.show_command('show', getter_name='get', table_transformer=transform_containerapp_output)

    with self.command_group('containerapp revision', containerapps_revisions_sdk, is_preview=True) as g:
        g.custom_command('show', 'get_revision', client_factory=cf_containerapps_revisions, table_transformer=transform_revision_output)
        g.custom_command('list', 'list_revisions', client_factory=cf_containerapps_revisions, table_transformer=transform_revision_list_output)
        g.custom_command('activate', 'activate_revision', client_factory=cf_containerapps_revisions)
        g.custom_command('deactivate', 'deactivate_revision', client_factory=cf_containerapps_revisions)
        g.custom_command('restart', 'restart_revision', client_factory=cf_containerapps_revisions)

    with self.command_group('containerapp github-action', containerapps_sdk, is_preview=True) as g:
        g.custom_command('add', 'create_or_update_github_action', client_factory=cf_containerapps, exception_handler=ex_handler_factory())
        g.custom_command('show', 'show_github_action', client_factory=cf_containerapps, exception_handler=ex_handler_factory())
        g.custom_command('delete', 'delete_github_action', client_factory=cf_containerapps, exception_handler=ex_handler_factory())

    # with self.command_group('containerapp secret', containerapps_sdk, is_preview=True) as g:
    #    g.custom_command('set', 'set_app_secret', client_factory=cf_containerapps, supports_no_wait=True)

    # with self.command_group('containerapp scale-rule', containerapps_sdk, is_preview=True) as g:
    #    g.custom_command('create', 'add_scale_rule', client_factory=cf_containerapps, supports_no_wait=True)
    #    g.custom_command('delete', 'remove_scale_rule', client_factory=cf_containerapps, supports_no_wait=True, confirmation=True)
    #    g.custom_command('update', 'update_scale_rule', client_factory=cf_containerapps)

    with self.command_group('containerapp env', kube_environments_sdk, is_preview=True) as g:
        g.show_command('show')
        g.wait_command('wait')
        g.custom_command('list', 'list_kube_environments', client_factory=cf_kube_environments)
        g.custom_command('create', 'create_kube_environment', client_factory=cf_kube_environments, supports_no_wait=True, exception_handler=ex_handler_factory())
        g.custom_command('update', 'update_kube_environment', client_factory=cf_kube_environments, supports_no_wait=True, exception_handler=ex_handler_factory())
        g.command('delete', 'delete', supports_no_wait=True, confirmation=True, exception_handler=ex_handler_factory())
