# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# import subprocess

from knack.util import CLIError
from knack.log import get_logger
from azure.cli.core.azclierror import (ResourceNotFoundError, RequiredArgumentMissingError, MutuallyExclusiveArgumentError, ValidationError)
from azure.cli.core.util import sdk_no_wait
from azure.cli.core.commands import LongRunningOperation
from azure.cli.core.commands.client_factory import get_subscription_id
from msrestazure.tools import is_valid_resource_id, parse_resource_id
from msrest.exceptions import DeserializationError
from ._client_factory import providers_client_factory
from urllib.parse import urlparse
import errno
import yaml

# from os import environ
# from os.path import join as path_join
from ._client_factory import web_client_factory, cf_resource_groups
from ._utils import scale_rules_from_file, parse_env_var_flags, parse_secret_flags, contains,\
                    validate_memory, validate_transport_method, validate_cpu, _get_location_from_resource_group,\
                    dapr_components_from_file, parse_traffic_weights_flags, parse_list_of_strings,\
                    combine_env_vars, combine_secrets, raise_missing_token_suggestion
from ._github_oauth import get_github_access_token
from azure.cli.command_modules.appservice.custom import (_get_acr_cred)

logger = get_logger(__name__)

def process_loaded_yaml(yaml_containerapp):
    if not yaml_containerapp.get('properties'):
        yaml_containerapp['properties'] = {}

    nested_properties = ['kubeEnvironmentId', 'latestRevisionFqdn', 'latestRevisionName', 'template', 'configuration', 'provisioningState']
    for nested_property in nested_properties:
        tmp = yaml_containerapp.get(nested_property)
        if tmp:
            yaml_containerapp['properties'][nested_property] = tmp
            del yaml_containerapp[nested_property]
    
    return yaml_containerapp

def load_yaml_file(file_name):
    try:
        with open(file_name) as stream:
            return yaml.safe_load(stream)
    except (IOError, OSError) as ex:
        if getattr(ex, 'errno', 0) == errno.ENOENT:
            raise CLIError('{} does not exist'.format(file_name))
        raise
    except (yaml.parser.ParserError, UnicodeDecodeError) as ex:
        raise CLIError('Error parsing {} ({})'.format(file_name, str(ex)))

def create_or_update_containerapp_yaml(cmd,
                                      client,
                                      name,
                                      resource_group_name,
                                      file_name,
                                      is_update,
                                      no_wait=False):
    yaml_containerapp = process_loaded_yaml(load_yaml_file(file_name))
    if type(yaml_containerapp) != dict:
        raise ValidationError('Invalid YAML provided. Please see https://github.com/microsoft/k4apps/wiki/WorkerApps-ARM-API-Specification for a valid containerapps YAML spec.')

    resource_group_name = yaml_containerapp.get('resourceGroup') if yaml_containerapp.get('resourceGroup') is not None else resource_group_name

    if not yaml_containerapp.get('name'):
        yaml_containerapp['name'] = name
    name = yaml_containerapp.get('name')

    if not yaml_containerapp.get('type'):
        yaml_containerapp['type'] = 'Microsoft.Web/ContainerApps'

    if not yaml_containerapp.get('kind'):
        yaml_containerapp['kind'] = 'containerapp'

    current_containerapp_def = None
    containerapp_def = None
    try:
        current_containerapp_def = client.get(resource_group_name, name)
    except Exception as ex:
        pass

    if not is_update and current_containerapp_def is not None:
        raise CLIError("The containerapp '{}' already exists. To update a containerapp please use the \"az containerapp update\" command.".format(name))
    elif is_update and  not current_containerapp_def:
        raise CLIError("The containerapp '{}' does not exist".format(name))

    try:
        containerapp_def = cmd.cli_ctx.deserializer('ContainerApp', yaml_containerapp)
    except DeserializationError as ex:
        raise ValidationError('Invalid YAML provided. Please see https://github.com/microsoft/k4apps/wiki/WorkerApps-ARM-API-Specification for a valid containerapps YAML spec.')

    # Validate parameters
    if containerapp_def.type.lower() != 'microsoft.web/containerapps':
        raise ValidationError('Containerapp type must be \"Microsoft.Web/ContainerApps\"')
    if containerapp_def.kind.lower() != 'containerapp':
        raise ValidationError('Containerapp kind must be \"containerapp\"')

    # Validate kube env
    if not containerapp_def.kube_environment_id:
        if is_update:
            containerapp_def.kube_environment_id = current_containerapp_def.kube_environment_id
        else:
            raise RequiredArgumentMissingError('kubeEnvironmentId is required. Please see https://github.com/microsoft/k4apps/wiki/WorkerApps-ARM-API-Specification for a valid containerapps YAML spec.')
    kube_env_id = containerapp_def.kube_environment_id
    if not kube_env_id:
        raise RequiredArgumentMissingError('kubeEnvironmentId is required. Please see https://github.com/microsoft/k4apps/wiki/WorkerApps-ARM-API-Specification for a valid containerapps YAML spec.')

    kube_name = None
    kube_rg = None
    kube_env_info = None

    if (is_valid_resource_id(kube_env_id)):
        parsed_kube_env = parse_resource_id(kube_env_id)
        kube_name = parsed_kube_env['name']
        kube_rg = parsed_kube_env['resource_group']
    else:
        raise ValidationError('Invalid kubeEnvironmentId specified. Environment not found')

    try:
        general_client = web_client_factory(cmd.cli_ctx)
        kube_env_info = general_client.kube_environments.get(kube_rg, kube_name)
    except:
        pass

    if not kube_env_info:
        raise CLIError("The environment '{}' does not exist".format(kube_env_id))

    kube_env_id = kube_env_info.id
    if not containerapp_def.location:
        containerapp_def.location = kube_env_info.location

    # Validate ActiveRevisionsMode
    if containerapp_def.configuration and containerapp_def.configuration.active_revisions_mode:
        if containerapp_def.configuration.active_revisions_mode.lower() not in ['single', 'multiple']:
            raise ValidationError('activeRevisionsMode is not valid. Value must be \"single\" or \"multiple\".')

    # Validate Ingress
    if containerapp_def.configuration and containerapp_def.configuration.ingress:
        if not containerapp_def.configuration.ingress.target_port:
            raise CLIError("Must specify 'target_port' for ingress")

        if containerapp_def.configuration.ingress.transport:
            if containerapp_def.configuration.ingress.transport.lower() not in ['auto', 'http']:
                raise ValidationError('transport is not valid. Value must be \"auto\" or \"http\"')

    if is_update:
        # Secrets
        existing_secrets = []
        try:
            existing_secrets = client.list_secrets(resource_group_name, name)
        except:
            pass

        if containerapp_def.configuration and containerapp_def.configuration.secrets:
            containerapp_def.configuration.secrets = combine_secrets(existing_secrets, containerapp_def.configuration.secrets)
        else:
            containerapp_def.configuration.secrets = existing_secrets

        # TODO: Control Plane should accept revisions_mode = single and traffic settings
        # traffic:
        # - latestRevision: true
        #   weight: 100
        if containerapp_def.configuration.active_revisions_mode.lower() == "single":
            if containerapp_def.configuration.ingress:
                containerapp_def.configuration.ingress.traffic = None

    return sdk_no_wait(no_wait, client.create_or_update, resource_group_name, name, container_app_envelope=containerapp_def)

def create_containerapp(cmd,
                        client,
                        name,
                        resource_group_name,
                        yaml=None,
                        image_name=None,
                        kube_env=None,
                        scale_rules=None,
                        min_replicas=None,
                        max_replicas=None,
                        target_port=None,
                        transport="auto",
                        ingress=None,
                        revisions_mode=None,
                        secrets=None,
                        env_vars=None,
                        cpu=None,
                        memory=None,
                        registry_server=None,
                        registry_user=None,
                        registry_pass=None,
                        dapr_enabled=False,
                        dapr_app_port=None,
                        dapr_app_id=None,
                        dapr_components=None,
                        location=None,
                        startup_command=None,
                        args=None,
                        tags=None,
                        no_wait=False):
    # Check if sub is registered to the RP
    try:
        providers_client = providers_client_factory(cmd.cli_ctx, get_subscription_id(cmd.cli_ctx))
        registration_state = getattr(providers_client.get('Microsoft.Web'), 'registration_state', "NotRegistered")

        if not (registration_state and registration_state.lower() == 'registered'):
            raise ValidationError('Subscription is not registered for the resource type. Please run \"az provider register -n Microsoft.Web --wait\" to register your subscription.')
    except ValidationError as ex:
        raise ex
    except Exception:
        pass

    if yaml:
        return create_or_update_containerapp_yaml(cmd=cmd, client=client, name=name, resource_group_name=resource_group_name, file_name=yaml, is_update=False, no_wait=no_wait)

    if image_name is None:
        raise RequiredArgumentMissingError('Usage error: --image is required if not using --yaml')

    if kube_env is None:
        raise RequiredArgumentMissingError('Usage error: --environment is required if not using --yaml')

    # Validate Parameters
    if cpu is not None:
        validate_cpu(cpu)
        if memory is None:
            raise RequiredArgumentMissingError("usage error: --memory required")
    elif memory is not None:
        validate_memory(memory)
        if cpu is None:
            raise RequiredArgumentMissingError("usage error: --cpu required")
    if ((registry_server is not None or registry_user is not None or registry_pass is not None) and 
          not(registry_server is not None and registry_user is not None and registry_pass is not None)):
        raise CLIError("usage error: --registry-login-server, --registry-password and --registry-username are required")
    if not validate_transport_method(transport):
        raise CLIError("usage error: --transport must be one of \"auto\", or \"http2\"")

    # Useful Resources
    ContainerApp, Template, Configuration, Ingress, Container,\
    ContainerResources, Dapr, Scale, RegistryCredentials, Secret =\
    cmd.get_models('ContainerApp', 'Template', 'Configuration', 'Ingress', 'Container',\
    'ContainerResources', 'Dapr', 'Scale', 'RegistryCredentials', 'Secret')

    containerapp_def = None
    try:
        containerapp_def = client.get(resource_group_name, name)
    except:
        pass

    if containerapp_def is not None:
        raise CLIError("The containerapp '{}' already exists. To update a containerapp please use the \"az containerapp update\" command.".format(name))
    
    kube_name = None
    kube_rg = None
    kube_env_info = None

    if (is_valid_resource_id(kube_env)):
        parsed_kube_env = parse_resource_id(kube_env)
        kube_name = parsed_kube_env['name']
        kube_rg = parsed_kube_env['resource_group']
    else:
        kube_name = kube_env
        kube_rg = resource_group_name

    try:
        general_client = web_client_factory(cmd.cli_ctx)
        kube_env_info = general_client.kube_environments.get(kube_rg, kube_name)
    except:
        pass
    
    if not kube_env_info:
        raise CLIError("The environment '{}' does not exist".format(kube_env))
    # elif location is not None and location != kube_env_info.location:
    #     raise CLIError("error: The location '{}' specified is different from environment '{}'\'s location '{}'".format(location, kube_env, kube_env_info.location))
    location = kube_env_info.location
    kube_env_id = kube_env_info.id

    external_ingress = True
    if ingress is not None:
        if ingress.lower() == "internal":
           external_ingress = False
        elif ingress.lower() == "external":
            external_ingress = True
 
    # App Configuration
    containerapp_ingress_def = None
    if ingress is not None and target_port is None:
        raise CLIError("Must specify 'target_port' for '{}' ingress".format(ingress))
    elif ingress is None and target_port is not None:
        raise CLIError("Must specify 'ingress' as either 'internal' or 'external'")
    elif target_port is not None and ingress is not None:
        containerapp_ingress_def = Ingress(external=external_ingress, target_port=target_port, transport=transport)

    containerapp_config_def = Configuration(ingress=containerapp_ingress_def)
    if secrets is not None:
        containerapp_secrets = parse_secret_flags(cmd, secrets)
        containerapp_config_def.secrets = containerapp_secrets
    
    if revisions_mode is not None:
        containerapp_config_def.active_revisions_mode = revisions_mode
    
    if registry_server is not None:
        containerapp_config_def.registries = []

        if not(registry_pass.startswith('secretref:')):
            if (urlparse(registry_server).hostname is not None):
                registry_secret_name = "{server}-{user}".format(server=urlparse(registry_server).hostname.replace('.', ''), user=registry_user.lower())
            else:
                registry_secret_name = "{server}-{user}".format(server=registry_server.replace('.', ''), user=registry_user.lower())
            
            if containerapp_config_def.secrets is None:
                containerapp_config_def.secrets = []
            
            containerapp_config_def.secrets.append(
                Secret(
                    name=registry_secret_name,
                    value=registry_pass
            ))
            containerapp_config_def.registries.append(
                RegistryCredentials(
                    server=registry_server,
                    username=registry_user,
                    password_secret_ref=registry_secret_name
            ))
        else:
            containerapp_config_def.registries.append(
                RegistryCredentials(
                    server=registry_server,
                    username=registry_user,
                    password_secret_ref=registry_pass
            ))

    # App Template
    containerapp_container_def = Container(image=image_name, name=name)
    
    if cpu is not None and memory is not None:
        containerapp_container_def.resources = ContainerResources(cpu=cpu, memory=memory)
    
    if env_vars is not None:
        containerapp_container_def.env = parse_env_var_flags(cmd, env_vars)

    if startup_command is not None:
        containerapp_container_def.command = parse_list_of_strings(startup_command)

    if args is not None:
        containerapp_container_def.args = parse_list_of_strings(args)

    containerapp_template_def = Template(containers=[containerapp_container_def])

    if min_replicas is not None or max_replicas is not None or scale_rules is not None:
        containerapp_scale_def = Scale()
        if max_replicas is not None:
            containerapp_scale_def.max_replicas = max_replicas
        if min_replicas is not None:
            containerapp_scale_def.min_replicas = min_replicas
        if scale_rules is not None:
            containerapp_scale_def.rules = scale_rules_from_file(cmd, scale_rules)
        
        containerapp_template_def.scale = containerapp_scale_def
        
    if dapr_enabled is True:
        containerapp_dapr_def = Dapr(enabled=dapr_enabled)
        if dapr_app_port is not None:
            containerapp_dapr_def.app_port = dapr_app_port
        if dapr_app_id is not None:
            containerapp_dapr_def.app_id = dapr_app_id
        if dapr_components is not None:
            containerapp_dapr_def.components = dapr_components_from_file(cmd, dapr_components)
        
        containerapp_template_def.dapr = containerapp_dapr_def
    

    containerapp_def = ContainerApp(name=name,
                              kind='containerapp',
                              type='Microsoft.Web/containerApps',
                              location=location,
                              kube_environment_id=kube_env_id,
                              template=containerapp_template_def,
                              configuration=containerapp_config_def)

    if tags:
        containerapp_def.tags = tags

    return sdk_no_wait(no_wait, client.create_or_update, resource_group_name, name, container_app_envelope=containerapp_def)

def list_containerapp(client, resource_group_name=None):
    if resource_group_name is None:
        return client.list_by_subscription()
    return client.list_by_resource_group(resource_group_name)

def update_containerapp(cmd,
                        client,
                        name,
                        resource_group_name,
                        yaml=None,
                        image_name=None,
                        scale_rules=None,
                        min_replicas=None,
                        max_replicas=None,
                        ingress=None,
                        target_port=None,
                        transport=None,
                        traffic_weights=None,
                        revisions_mode=None,
                        secrets=None,
                        env_vars=None,
                        cpu=None,
                        memory=None,
                        registry_server=None,
                        registry_user=None,
                        registry_pass=None,
                        dapr_enabled=None,
                        dapr_app_port=None,
                        dapr_app_id=None,
                        dapr_components=None,
                        startup_command=None,
                        args=None,
                        tags=None,
                        no_wait=False):
    if yaml:
        return create_or_update_containerapp_yaml(cmd=cmd, client=client, name=name, resource_group_name=resource_group_name, file_name=yaml, is_update=True, no_wait=no_wait)

    containerapp_def = None
    try:
        containerapp_def = client.get(resource_group_name, name)
    except:
        pass
    
    if not containerapp_def:
        raise CLIError("The containerapp '{}' does not exist".format(name))

    update_map = {}
    update_map['secrets'] = secrets is not None
    update_map['ingress'] = ingress or target_port or transport or traffic_weights
    update_map['registries'] = registry_server or registry_user or registry_pass
    update_map['scale'] = None not in (min_replicas, max_replicas, scale_rules)
    update_map['container'] = None not in (image_name, env_vars, cpu, memory, startup_command, args)
    update_map['dapr'] = None not in (dapr_enabled, dapr_app_port, dapr_app_id, dapr_components)
    update_map['configuration'] = update_map['secrets'] or update_map['ingress'] or update_map['registries'] or revisions_mode is not None

    # Validations
    if cpu is not None:
        validate_cpu(cpu)
    elif memory is not None:
        validate_memory(memory)
    if transport is not None and not validate_transport_method(transport):
        raise CLIError("usage error: --transport must be one of \"auto\", or \"http2\"")
    if containerapp_def is None:
        raise CLIError("The containerapp '{}' does not exist".format(name))
    if (update_map['registries'] and 
          not(registry_server is not None and registry_user is not None and registry_pass is not None)):
        raise CLIError("usage error: --registry-login-server, --registry-password and --registry-username are required")
    if update_map['container'] and len(containerapp_def.template.containers) > 1:
        raise CLIError("usage error: trying to update image, environment variables, resources claims on a multicontainer containerapp. Please use arm templates for multicontainer containerapp update")
    
    # Useful Resources
    Configuration, Ingress, ContainerResources,\
    Dapr, Scale, RegistryCredentials, Secret =\
    cmd.get_models('Configuration', 'Ingress', 'ContainerResources',\
    'Dapr', 'Scale', 'RegistryCredentials', 'Secret')

    if tags:
        containerapp_def.tags = tags

    # Containers
    if image_name is not None:
        containerapp_def.template.containers[0].image = image_name
    if env_vars is not None:
        containerapp_def.template.containers[0].env = combine_env_vars(containerapp_def.template.containers[0].env, parse_env_var_flags(cmd, env_vars, is_update_containerapp=True))
    if startup_command is not None:
        containerapp_def.template.containers[0].command = parse_list_of_strings(startup_command)
    if args is not None:
        containerapp_def.template.containers[0].args = parse_list_of_strings(args)
    if cpu is not None or memory is not None:
        if containerapp_def.template.containers[0].resources is None:
            if cpu is not None:
                if memory is None:
                    raise CLIError("usage error: --memory required")
            elif memory is not None:
                if cpu is None:
                    raise CLIError("usage error: --cpu required")
            containerapp_def.template.containers[0].resources = ContainerResources(cpu=cpu, memory=memory)
        else:
            if cpu is not None:
                containerapp_def.template.containers[0].resources.cpu = cpu
            if memory is not None:
                containerapp_def.template.containers[0].resources.memory = memory

    # Scale
    if update_map['scale']:
        if containerapp_def.template.scale is None:
            containerapp_def.template.scale = Scale()
    if min_replicas is not None:
        containerapp_def.template.scale.min_replicas = min_replicas
    if max_replicas is not None:
        containerapp_def.template.scale.max_replicas = max_replicas
    if scale_rules is not None:
        containerapp_def.template.scale.rules = scale_rules_from_file(cmd, scale_rules)

    # Dapr
    if update_map['dapr']:
        if containerapp_def.template.dapr is None:
            containerapp_def.template.dapr = Dapr()
        if dapr_enabled is not None:
            containerapp_def.template.dapr.enabled = dapr_enabled
        if dapr_app_port is not None:
            containerapp_def.template.dapr.app_port = dapr_app_port
        if dapr_app_id is not None:
            containerapp_def.template.dapr.app_id = dapr_app_id
        if dapr_components is not None:
            containerapp_def.template.dapr.components = dapr_components_from_file(cmd, dapr_components)

    # Configuration
    if update_map['configuration']:
        if containerapp_def.configuration is None:
            containerapp_def.configuration = Configuration()

        if revisions_mode is not None:
            containerapp_def.configuration.active_revisions_mode = revisions_mode
        
        if update_map['ingress']:
            if not containerapp_def.configuration.ingress:
                containerapp_def.configuration.ingress = Ingress()

            if ingress and ingress.lower() == "none":
                containerapp_def.configuration.ingress = None
            else:
                if ingress is not None:
                    # No existing port for the app and no port provided as a part of the update call
                    if target_port is None and containerapp_def.configuration.ingress.target_port is None:
                        raise CLIError("Must specify 'target_port' for ingress '{}'".format(ingress))
                    if ingress.lower() == "external":
                        containerapp_def.configuration.ingress.external = True
                    elif ingress.lower() == "internal":
                        containerapp_def.configuration.ingress.external = False
                if target_port is not None:
                    if ingress is None and containerapp_def.configuration.ingress is None:
                        raise CLIError("Must specify 'ingress' as either 'internal' or 'external'")
                    containerapp_def.configuration.ingress.target_port = target_port
                if transport is not None:
                    containerapp_def.configuration.ingress.transport = transport
                if traffic_weights is not None:
                    containerapp_def.configuration.ingress.traffic = parse_traffic_weights_flags(cmd, traffic_weights)

        if update_map['registries']:
            containerapp_def.configuration.registries = []

            if not(registry_pass.startswith('secretref:')):
                if (urlparse(registry_server).hostname is not None):
                    registry_secret_name = "{server}-{user}".format(server=urlparse(registry_server).hostname.replace('.', ''), user=registry_user.lower())
                else:
                    registry_secret_name = "{server}-{user}".format(server=registry_server.replace('.', ''), user=registry_user.lower())
                
                if containerapp_def.configuration.secrets is None:
                    containerapp_def.configuration.secrets = []
                
                # Add registry password as a secret with name "{container registry}-{container registry user}"
                if not secrets:
                    secrets = ""
                if secrets and secrets[-1] != ',':
                    secrets += ","
                secrets += registry_secret_name
                secrets += "="
                secrets += registry_pass
                update_map['secrets'] = True

                containerapp_def.configuration.registries.append(
                    RegistryCredentials(
                        server=registry_server,
                        username=registry_user,
                        password_secret_ref=registry_secret_name
                ))
            else:
                containerapp_def.configuration.registries.append(
                    RegistryCredentials(
                        server=registry_server,
                        username=registry_user,
                        password_secret_ref=registry_pass
                ))

    # Secrets
    existing_secrets = []
    try:
        existing_secrets = client.list_secrets(resource_group_name, name)
    except:
        pass

    if update_map['secrets']:
        containerapp_def.configuration.secrets = combine_secrets(existing_secrets, parse_secret_flags(cmd, secrets))
    else:
        containerapp_def.configuration.secrets = existing_secrets

    # TODO: Control Plane should accept revisions_mode = single and traffic settings 
    # traffic:
    # - latestRevision: true
    #   weight: 100
    if containerapp_def.configuration.active_revisions_mode.lower() == "single":
        if containerapp_def.configuration.ingress:
            containerapp_def.configuration.ingress.traffic = None

    if containerapp_def.template.revision_suffix is not None:
        containerapp_def.template.revision_suffix = None

    return sdk_no_wait(no_wait, client.create_or_update, resource_group_name, name, containerapp_def)

def delete_containerapp(client, name, resource_group_name, no_wait=False):
    containerapp_def = client.get(resource_group_name, name)
    if containerapp_def is None:
        raise CLIError("The containerapp '{}' does not exist".format(name))

    return sdk_no_wait(no_wait, client.delete, resource_group_name, name)

def list_revisions(client, name, resource_group_name):
    resp = client.list_revisions(resource_group_name, name)
    return resp

def get_revision(client, name, app, resource_group_name):
    resp = client.get_revision(resource_group_name, app, name)
    return resp

def activate_revision(client, name, app, resource_group_name):
    client.activate_revision(resource_group_name, app, name)
    print("Revision activated successfully")
    return

def deactivate_revision(client, name, app, resource_group_name):
    client.deactivate_revision(resource_group_name, app, name)
    print("Revision deactivated successfully")
    return

def restart_revision(client, name, app, resource_group_name):
    client.restart_revision(resource_group_name, app, name)
    print("Revision instances restarted successfully")
    return

# TODO: This method should be removed Secrets should be added through azure portal or on creation with yaml
def set_app_secret(cmd,
                   client,
                   resource_group_name,
                   name,
                   secret_name,
                   value,
                   no_wait=False):
    # Useful Resources
    Secret = cmd.get_models('Secret')

    # Get App Information
    containerapp_def = client.get(resource_group_name, name)
    if containerapp_def is None:
        raise CLIError("The containerapp '{}' does not exist".format(name))
    
    if len(containerapp_def.template.containers) > 1:
        raise CLIError("Please update port using yaml config update")
    elif len(containerapp_def.template.containers) < 1:
        raise CLIError("Container App has no container")

    secret_def = Secret(name=secret_name, value=value)

    if containerapp_def.configuration.secrets is None:
        containerapp_def.configuration.secrets = [secret_def]
    else:
        containerapp_def.configuration.secrets.append(secret_def)

    return sdk_no_wait(no_wait, client.create_or_update, resource_group_name, name, containerapp_def)

def set_app_environment_var(cmd,
                       client,
                       resource_group_name,
                       name,
                       value,
                       secret_ref=False,
                       no_wait=False):
    # Useful Resources
    EnvironmentVar = cmd.get_models('EnvironmentVar')

    # Get App Information
    containerapp_def = client.get(resource_group_name, name)
    if containerapp_def is None:
        raise CLIError("The containerapp '{}' does not exist".format(name))

    if len(containerapp_def.template.containers) > 1:
        raise CLIError("Please update port using yaml config update")
    elif len(containerapp_def.template.containers) < 1:
        raise CLIError("Container App has no container")

    env_pair_strings = value.split(',')
    env_pairs = {}

    for pair in env_pair_strings:
        key_val = pair.split('=')
        if len(key_val) is not 2:
            raise CLIError("usage error: --value must be in format \"<key>=<value>,<key>=<value>,...\"")
        env_pairs[key_val[0]] = key_val[1]

    if containerapp_def.template.containers[0].env is None or len(containerapp_def.template.containers[0].env) is 0:
        env_var_def = []
        for key, value in env_pairs.items():
            if secret_ref is True:
                env_var_def.append(EnvironmentVar(name=key, secret_ref=value))
            else:
                env_var_def.append(EnvironmentVar(name=key, value=value))
        containerapp_def.template.containers[0].env = env_var_def
    else:
        env_var_def = []
        # If environment variable exists in app config then update

        for i in range(len(containerapp_def.template.containers[0].env)):
            key = containerapp_def.template.containers[0].env[i].name
            if key in env_pairs:
                if secret_ref is True:
                    containerapp_def.template.containers[0].env[i] = EnvironmentVar(name=key, secret_ref=env_pairs[key])
                else:
                    containerapp_def.template.containers[0].env[i] = EnvironmentVar(name=key, value=env_pairs[key])
                env_pairs.pop(key)

        for key, value in env_pairs.items():
            if secret_ref is True:
                env_var_def.append(EnvironmentVar(name=key, secret_ref=value))
            else:
                env_var_def.append(EnvironmentVar(name=key, value=value))

        # Add non existent environment variables to app config
        containerapp_def.template.containers[0].env = containerapp_def.template.containers[0].env + env_var_def

        containerapp_def.configuration.secrets = None

    return sdk_no_wait(no_wait, client.create_or_update, resource_group_name, name, containerapp_def)

def scale_containerapp(cmd, client, resource_group_name, name, min_replicas, max_replicas, no_wait=False):
    logger.warning('Warning: scale command may be removed in the upcoming releases')
    # Useful Resources
    Scale = cmd.get_models('Scale')

    # Get App Information
    containerapp_def = client.get(resource_group_name, name)
    if containerapp_def is None:
        raise CLIError("The containerapp '{}' does not exist".format(name))
    if containerapp_def.template.scale is None:
        containerapp_def.template.scale = Scale(min_replicas=min_replicas, max_replicas=max_replicas)

    containerapp_def.template.scale.min_replicas = min_replicas
    containerapp_def.template.scale.max_replicas = max_replicas
    containerapp_def.configuration.secrets = None

    return sdk_no_wait(no_wait, client.create_or_update, resource_group_name, name, containerapp_def)

def add_scale_rule(cmd,
                   client,
                   resource_group_name,
                   name,
                   rule_name,
                   rule_type=None,
                   queue_name=None,
                   queue_length=None,
                   metadata=None,
                   custom_type=None,
                   secret_ref=None,
                   trigger_parameter=None,
                   no_wait=False):
    # Validate Parameters
    if rule_type is None:
        raise CLIError("usage error: --type required")
    elif rule_type == 'azure-queue':
        if queue_name is None:
            raise CLIError("usage error: --queue-name required")
        if queue_length is None:
            raise CLIError("usage error: --queue-length required")
    elif rule_type == 'custom':
        if metadata is None:
            raise CLIError("usage error: --metadata required")
        if custom_type is None:
            raise CLIError("usage error: --custom-type required")
    elif rule_type == 'http':
        if metadata is None:
            raise CLIError("usage error: --metadata required")
    else:
        raise CLIError("usage error: --rule-type needs to be one of azure-queue, custom or http")

    # Useful Resources
    ScaleRule, QueueScaleRule, ScaleRuleAuth, \
    CustomScaleRule, HttpScaleRule, Scale = \
    cmd.get_models('ScaleRule',  'QueueScaleRule', 'ScaleRuleAuth',\
                   'CustomScaleRule', 'HttpScaleRule', 'Scale')

    # Get App Information
    containerapp_def = client.get(resource_group_name, name)
    if containerapp_def is None:
        raise CLIError("The containerapp '{}' does not exist".format(name))

    auth = False
    if secret_ref is None and trigger_parameter is not None:
        auth = True
        containerapp_scale_auth_def = ScaleRuleAuth(secret_ref=secret_ref, trigger_parameter=trigger_parameter)
    elif secret_ref is None and trigger_parameter is not None or secret_ref is not None and trigger_parameter is None:
        raise CLIError("usage error: --secret-ref and --trigger required")
    
    if rule_type == 'azure-queue':
        if auth is True:
            containerapp_queue_rule_def = QueueScaleRule(queue_name=queue_name, queue_length=queue_length, auth=containerapp_scale_auth_def)
        else:
            containerapp_queue_rule_def = QueueScaleRule(queue_name=queue_name, queue_length=queue_length)
        containerapp_rule_def = ScaleRule(name=rule_name, azure_queue=containerapp_queue_rule_def)
    elif rule_type == 'custom':
        if auth is True:
            containerapp_custom_rule_def = CustomScaleRule(type=custom_type, metadata=metadata, auth=containerapp_scale_auth_def)
        else:
            containerapp_custom_rule_def = CustomScaleRule(type=custom_type, metadata=metadata)
        containerapp_rule_def = ScaleRule(name=rule_name, custom=containerapp_custom_rule_def)
    elif rule_type == 'http':
        if auth is True:
            containerapp_http_rule_def = HttpScaleRule(metadata=metadata, auth=containerapp_scale_auth_def)
        else:
            containerapp_http_rule_def = HttpScaleRule(metadata=metadata)
        containerapp_rule_def = ScaleRule(name=rule_name, http=containerapp_http_rule_def)

    if containerapp_def.template.scale is None:
        containerapp_def.template.scale = Scale(rules=[containerapp_rule_def])
    elif containerapp_def.template.scale.rules is None:
        containerapp_def.template.scale.rules = [containerapp_rule_def]
    else:
        containerapp_def.template.scale.rules.append(containerapp_rule_def)

    containerapp_def.configuration.secrets = None

    return sdk_no_wait(no_wait, client.create_or_update, resource_group_name, name, containerapp_def)

def remove_scale_rule(client,
                      resource_group_name,
                      name,
                      rule_name,
                      no_wait=False):

    # Get App Information
    containerapp_def = client.get(resource_group_name, name)
    if containerapp_def is None:
        raise CLIError("The containerapp '{}' does not exist".format(name))

    if containerapp_def.template.scale is None:
        raise CLIError("The scale rule '{}' does not exist".format(rule_name))

    rule_exists = False
    for i in range(0, len(containerapp_def.template.scale.rules)):
        if containerapp_def.template.scale.rules[i].name == rule_name:
            rule_exists = True
            del containerapp_def.template.scale.rules[i]
            break

    if rule_exists is False:
        raise CLIError("The scale rule '{}' does not exist".format(rule_name))

    containerapp_def.configuration.secrets = None

    return sdk_no_wait(no_wait, client.create_or_update, resource_group_name, name, containerapp_def)

def update_scale_rule():
    # raise CLIError('TODO: Implement `containerapp scale-rule update`')
    return

def create_kube_environment(cmd,
                            client,
                            name,
                            resource_group_name,
                            logs_customer_id,
                            logs_key,
                            logs_destination="log-analytics",
                            location=None,
                            instrumentation_key=None,
                            controlplane_subnet_resource_id=None,
                            app_subnet_resource_id=None,
                            tags=None,
                            no_wait=False):
    # Check if sub is registered to the RP
    try:
        providers_client = providers_client_factory(cmd.cli_ctx, get_subscription_id(cmd.cli_ctx))
        registration_state = getattr(providers_client.get('Microsoft.Web'), 'registration_state', "NotRegistered")

        if not (registration_state and registration_state.lower() == 'registered'):
            raise ValidationError('Subscription is not registered for the resource type. Please run \"az provider register -n Microsoft.Web --wait\" to register your subscription.')
    except ValidationError as ex:
        raise ex
    except Exception:
        pass

    KubeEnvironment, ContainerAppsConfiguration, AppLogsConfiguration, LogAnalyticsConfiguration = cmd.get_models('KubeEnvironment', 'ContainerAppsConfiguration', 'AppLogsConfiguration', 'LogAnalyticsConfiguration')
    location = location or _get_location_from_resource_group(cmd.cli_ctx, resource_group_name)

    containerapps_config_def = ContainerAppsConfiguration()

    if instrumentation_key is not None:
        containerapps_config_def.dapr_a_i_instrumentation_key = instrumentation_key

    if controlplane_subnet_resource_id is not None:
        containerapps_config_def.control_plane_subnet_resource_id = controlplane_subnet_resource_id

    if app_subnet_resource_id is not None:
        containerapps_config_def.app_subnet_resource_id = app_subnet_resource_id

    app_logs_config_def = AppLogsConfiguration(
        destination=logs_destination,
        log_analytics_configuration=LogAnalyticsConfiguration(
            customer_id=logs_customer_id,
            shared_key=logs_key
        )
    )

    kube_def = KubeEnvironment(
        location=location,
        internal_load_balancer_enabled=False,
        kube_environment_type='managed',
        type="managed",
        app_logs_configuration=app_logs_config_def,
        container_apps_configuration=containerapps_config_def,
        tags=tags)

    # TODO Add cross tenant vnet logic
    # myenv = environ.copy()
    # if 'VIRTUAL_ENV' in environ:
    #     myenv['PATH'] = ':'.join(
    #         [x for x in environ['PATH'].split(':')
    #             if x != path_join(environ['VIRTUAL_ENV'], 'bin')])
    
    # create_service_principle_process = subprocess.Popen(['az', 'ad', 'sp', 'create', '--id', APP_KEY_HERE], env=myenv)

    # create_rbac_process = subprocess.Popen(['az', 'role', 'assignment', 'create', '--assignee', APP_KEY_HERE, '--scope', SUBNET_RESOURCE_ID_HERE, '--role', '"Network Contributor"'], env=myenv)

    return sdk_no_wait(no_wait, client.create_or_update, resource_group_name, name, kube_def)
    

def list_kube_environments(client, resource_group_name=None):
    kube_env_list = None

    if resource_group_name is None:
        kube_env_list = client.list_by_subscription()
    else:
        kube_env_list = client.list_by_resource_group(resource_group_name)
    return [e for e in kube_env_list if e.kube_environment_type == "managed"]


def update_kube_environment(cmd,
                            kube_name,
                            resource_group_name,
                            tags=None,
                            no_wait=False):
    raise CLIError("Update is not yet supported for Kubernetes Environments.")


def create_or_update_github_action(cmd,
                                   client,
                                   name,
                                   resource_group_name,
                                   repo_url,
                                   registry_url=None,
                                   registry_username=None,
                                   registry_password=None,
                                   branch=None,
                                   token=None,
                                   login_with_github=False,
                                   docker_file_path=None,
                                   service_principal_client_id=None,
                                   service_principal_client_secret=None,
                                   service_principal_tenant_id=None):
    if not token and not login_with_github:
        raise_missing_token_suggestion()
    elif not token:
        scopes = ["admin:repo_hook", "repo", "workflow"]
        token = get_github_access_token(cmd, scopes)
    elif token and login_with_github:
        logger.warning("Both token and --login-with-github flag are provided. Will use provided token")

    try:
        # Verify github repo
        from github import Github, GithubException
        from github.GithubException import BadCredentialsException, UnknownObjectException

        repo = None
        repo = repo_url.split('/')
        if len(repo) >= 2:
            repo = '/'.join(repo[-2:])
        
        if repo:
            g = Github(token)
            github_repo = None
            try:
                github_repo = g.get_repo(repo)
                try:
                    github_repo.get_branch(branch=branch)
                except GithubException as e:
                    error_msg = "Encountered GitHub error when accessing {} branch in {} repo.".format(branch, repo)
                    if e.data and e.data['message']:
                        error_msg += " Error: {}".format(e.data['message'])
                    raise CLIError(error_msg)
                logger.warning('Verified GitHub repo and branch')
            except BadCredentialsException:
                raise CLIError("Could not authenticate to the repository. Please create a Personal Access Token and use "
                            "the --token argument. Run 'az webapp deployment github-actions add --help' "
                            "for more information.")
            except GithubException as e:
                error_msg = "Encountered GitHub error when accessing {} repo".format(repo)
                if e.data and e.data['message']:
                    error_msg += " Error: {}".format(e.data['message'])
                raise CLIError(error_msg)
    except CLIError as clierror:
        raise clierror
    except Exception as ex:
        # If exception due to github package missing, etc just continue without validating the repo and rely on api validation
        pass

    SourceControlInfo, GithubActionConfigurationContainerapp, AzureCredentials, RegistryInfo, SourceControlInfoApiResourceEnvelope = cmd.get_models(
        'SourceControlInfo', 'GithubActionConfigurationContainerapp', 'AzureCredentials', 'RegistryInfo', 'SourceControlInfoApiResourceEnvelope')
    source_control_info = None

    try:
        source_control_info = client.get_source_control_info(resource_group_name, name).properties
    except Exception as ex:
        if not service_principal_client_id or not service_principal_client_secret or not service_principal_tenant_id:
            raise RequiredArgumentMissingError('Service principal client ID, secret and tenant ID are required to add github actions for the first time. Please create one using the command \"az ad sp create-for-rbac --name \{name\} --role contributor --scopes /subscriptions/\{subscription\}/resourceGroups/\{resourceGroup\} --sdk-auth\"')
        source_control_info = SourceControlInfo()

    source_control_info.repo_url = repo_url

    if branch:
        source_control_info.branch = branch
    elif not source_control_info.branch:
        source_control_info.branch = "master"

    azure_credentials = None
    if service_principal_client_id or service_principal_client_secret or service_principal_tenant_id:
        azure_credentials = AzureCredentials(client_id=service_principal_client_id,
                                             client_secret=service_principal_client_secret,
                                             tenant_id=service_principal_tenant_id,
                                             subscription_id=get_subscription_id(cmd.cli_ctx))

    # Registry
    if not registry_username or not registry_password:
        # If registry is Azure Container Registry, we can try inferring credentials
        if not registry_url or '.azurecr.io' not in registry_url:
            raise RequiredArgumentMissingError('Registry url is required if using Azure Container Registry, otherwise Registry username and password are required if using Dockerhub')
        logger.warning('No credential was provided to access Azure Container Registry. Trying to look up...')
        parsed = urlparse(registry_url)
        registry_name = (parsed.netloc if parsed.scheme else parsed.path).split('.')[0]

        try:
            registry_username, registry_password = _get_acr_cred(cmd.cli_ctx, registry_name)
        except Exception as ex:
            raise RequiredArgumentMissingError('Failed to retrieve credentials for container registry. Please provide the registry username and password')

    registry_info = RegistryInfo(registry_url=registry_url,
                                 registry_user_name=registry_username,
                                 registry_password=registry_password)

    github_action_configuration = GithubActionConfigurationContainerapp(registry_info=registry_info,
                                                                        azure_credentials=azure_credentials,
                                                                        dockerfile_path=docker_file_path)

    source_control_info.github_action_configuration = github_action_configuration

    headers = {
        "x-ms-github-auxiliary": token
    }

    body = SourceControlInfoApiResourceEnvelope(properties=source_control_info)
    return client.create_or_update_source_control_info(resource_group_name, name, source_control_info=body, custom_headers=headers)


def show_github_action(cmd, client, name, resource_group_name):
    return client.get_source_control_info(resource_group_name, name)


def delete_github_action(cmd, client, name, resource_group_name, token=None, login_with_github=False):
    if not token and not login_with_github:
        raise_missing_token_suggestion()
    elif not token:
        scopes = ["admin:repo_hook", "repo", "workflow"]
        token = get_github_access_token(cmd, scopes)
    elif token and login_with_github:
        logger.warning("Both token and --login-with-github flag are provided. Will use provided token")

    headers = {
        "x-ms-github-auxiliary": token
    }

    return client.delete_source_control_info(resource_group_name, name, custom_headers=headers)
